/*
DL2 Monitor: ./pa/paip.c
Copyright (C) 2002, Rolf Jger, Ulm (Germany)

See the file README for information on usage and redistribution
of this source code file, and for a DISCLAIMER OF ALL WARRANTIES.
*/

/*----------------------------------------------------------------------------
 * Datalink Monitor (paip): Protocol Analysis (RFC791)
 * Internet Protocol, Version 4 (IPv4)
 *
 * Implementation Notes
 * default contains: length, flags, protocol, destination-address
 * 
 * Author: Rolf Jger,Ulm
 *----------------------------------------------------------------------------
 * YYYYMMDD  History
 * 20020501  icmp layer introduced and allowed.
 * 20020427  udp layer introduced, data and follow_* implemented.
 * 20020414  tcp layer introduced.
 * 20020405  first implementation without follow_frame.
 * 20020330  created from scratch
 * $Id: paip.c,v 1.6 2002/05/01 10:19:07 dv Exp $
 *----------------------------------------------------------------------------
*/
# include <stdio.h>
# include <string.h>
# include "lib/dl2pa.h"
# include "lib/paip.h"
# include "pa/ip.h"

// IPv4 frame positions:
# define P_VERSION 0x00
# define P_SERVICE_TYPE 0x01
# define P_LENGTH 0x02
# define P_FRAGMENT_ID 0x04
# define P_FLAGS 0x06
# define P_FRAGMENT_OFFS P_FLAGS
# define P_TIME_TO_LIVE 0x08
# define P_PROTOCOL 0x09
# define P_CHECKSUM 0x0a
# define P_SOURCE_ADDR 0x0c
# define P_DESTINATION_ADDR 0x10
# define P_OPTIONS 0x14
# define P_DATA hlength
# define MIN_HLENGTH P_OPTIONS

//-- common to any IP frame --------------------------------------------------
typedef struct Ipinfo {
  Octet protocol; 
  const char *name;
  int pid;
  int (*analyzer)(Dl2_Frame*,int);
  int follow_mask;
} Ipinfo;

# define IPINFO_ADD(t,a) {PAIP_T_##t,#t,PA_PID_##t,a,IP_Mfollow_##t}
static const Ipinfo ipinfo[] = {
  IPINFO_ADD(tcp,dl2_patcp),
  IPINFO_ADD(udp,dl2_paudp),
  IPINFO_ADD(icmp,dl2_paicmp),
  IPINFO_ADD(ip,0), 
  {0,NULL,0,0} 
};

static const Ipinfo* get_ipinfo(Octet);

//-exports--------------------------------------------------------------------
int dl2_paip(Dl2_Frame *frame,int verbose) {
  Octet *buffer = frame->buffer;
  int length = frame->length, hlength, pl, i = P_VERSION;
  Octet protocol;
  const Ipinfo *pi;

  // check for required packet length:
  if (length < MIN_HLENGTH) return -1;

  // compute header-length, data-length:
  hlength = (buffer[P_VERSION] & 0xf)*4;
  pl = str2doct(buffer+P_LENGTH)-hlength;
  if (pl > length) pl = length;

  // version:
  if (verbose & IP_Mversion) {
    printf(" %s=%01x",IP_Fversion,buffer[i]>>4);
  }

  // header-length:
  if (verbose & IP_Mheader_length) {
    printf(" %s=%01x",IP_Fheader_length,buffer[i]&0xf);
  } i = P_SERVICE_TYPE; 
  
  // service-type:
  if (verbose & IP_Mservice_type) {
    printf(" %s=%02x",IP_Fservice_type,buffer[i]);
  } i = P_LENGTH;
 
  // length:
  if (verbose & (IP_Mlength|IP_Mdefault)) {
    int dlength = str2doct(buffer+P_LENGTH)-hlength;
    printf(" %s=%04x",IP_Flength,dlength);
  } i = P_FRAGMENT_ID;

  // fragment-id:
  if (verbose & IP_Mfragment_id) {
    printf(" %s=",IP_Ffragment_id);
    for (;i < P_FLAGS; i++) printf("%02x",buffer[i]);
  } else i = P_FLAGS;

  // flags:
  if (verbose & (IP_Mflags|IP_Mdefault)) {
    int d = buffer[P_FLAGS] >>5;
    printf(" %s=%d%d",IP_Fflags,d>>1,d&1);
  } i = P_FRAGMENT_OFFS;

  // fragment-offset:
  if (verbose & IP_Mfragment_offset) {
    Doctet fo = str2doct(buffer+P_FRAGMENT_OFFS) & 0x1fff;
    printf(" %s=%04x",IP_Ffragment_offset,fo);
  } i = P_TIME_TO_LIVE;

  // time-to-live:
  if (verbose & IP_Mtime_to_live) {
    printf(" %s=%02x",IP_Ftime_to_live,buffer[P_TIME_TO_LIVE]);
  } i = P_PROTOCOL;

  // protocol:
  protocol = buffer[P_PROTOCOL];
  pi = get_ipinfo(protocol);
  if (verbose & IP_Mprotocol) {
    printf(" %s=%02x",IP_Fprotocol,protocol);
  }
  else if (verbose & IP_Mdefault) {
    if (pi != NULL) printf(" %s",pi->name);
  } i = P_CHECKSUM;
 
  // checksum:
  if (verbose & IP_Mchecksum) {
    printf(" %s=",IP_Fchecksum);
    for (;i < P_SOURCE_ADDR; i++) printf("%02x",buffer[i]);
  } else i = P_SOURCE_ADDR;
 
  // source-address:
  if (verbose & IP_Msource_addr) {
    printf(" %s=",IP_Fsource_addr);
    for (;i < P_DESTINATION_ADDR; i++) printf("%02x",buffer[i]);
  } else i = P_DESTINATION_ADDR;

  // destination-address: 
  if (verbose & (IP_Mdestination_addr|IP_Mdefault)) {
    printf(" %s=",IP_Fdestination_addr);
    for (;i < P_OPTIONS; i++) printf("%02x",buffer[i]);
  } else i = P_OPTIONS;

  // options: tbd
  i = P_DATA;
  frame->length = pl;
  frame->buffer = buffer+P_DATA;

  // data:
  if (verbose & IP_Mdata) {
    printf(" %s=",IP_Fdata);
    for (;i < length; i++) printf("%02x",buffer[i]);
  }

  // follow-frame:
  else if (pi != NULL && pi->analyzer) {
    int j;
    if ((verbose & IP_Mfollow_frame) ||
        (verbose & pi->follow_mask)) {
      printf(" [");
      j = pi->analyzer(frame,pa_verbose(pi->pid));
      printf(" ]");
      i += j;
    }
  } else i += pl;
  return i;
} //dl2_paip


//-locals---------------------------------------------------------------------
const Ipinfo* get_ipinfo(Octet protocol) {
  const Ipinfo *pi;
  for (pi = ipinfo; pi->protocol != 0; pi++) {
    if (pi->protocol == protocol) return pi;
  } return NULL;
} //get_ipinfo

/***EOM***/
